local M = {}

local scenario = require("scenario/scenarios")
local helper = require("scenario/scenariohelper")

local playerName = "rla_player0"
local trafficNames = {"rla_traffic0", "rla_traffic1", "rla_traffic2", "rla_traffic3", "rla_traffic4"}
local barrierNames = {"rla_barrier0", "rla_barrier1"}
local failMessages = {busted = "Busted! Step out of the vehicle, now!", crashBusted = "Totaled... and busted!", crash = "Totaled! Ouch.", overTime = "You ran out of time! What a disappointment."}
local winMessages = {normal = "Yay, you made it! The managers saw your vehicle and will keep the doors open for you.", lawful = "Yay, you made it! Nicely done, you didn't upset any cops!", lawless = "Yay, you made it! You took care of those roadblocks like a rock star.", veryFast = "Yay, you made it! That was fast! I admire your driving skills.", ultraFast = "Yay, you made it! Wow, that was insanely fast! Some of the local racers were informed, and they'd like to recruit you. That's crazy.", almost = "Yay, you made it! You got there with only seconds to spare!", policeCar = "Yay, you made it! Great vehicle choice; that's one way to deal with traffic!", funnyCar = "Yay, you made it! Oh wow, with that vehicle? Looks like you really know how to get around! Bravo.", evilCar = "Yay, you made it! That was pretty evil, though. What kind of monster are you?"}
local outMessages = {roadClosed = "Hey, you can't go through the construction zone!", wrongWay = "Wrong way! Nice try, buddy.", hospital = "You went to the hospital? I hope you're alright..."}
local paintColors = {{{1, 1, 1, 1.4}}, {{0, 0, 0, 1.4}, {0.5, 0, 0, 1.4}, {0.33, 0.33, 0.33, 1.4}, {0.0, 0.1, 0.42, 1.4}, {0.3, 0.2, 0.12, 1.4}, {0.83, 0.28, 0.03, 1.4}}, {{0.0, 0.05, 0.25, 1.2}, {0.5, 0, 0, 1.2}, {0.18, 0.18, 0.18, 1.2}, {0.45, 0.28, 0.233, 1.3}, {1, 0, 0, 1.2}, {0.22, 0.37, 0.33, 1.3}}, {{0.9, 0.05, 0.04, 1.35}, {1, 0.3, 0, 1.3}, {0.95, 0.7, 0.1, 1.35}, {0.0, 0.2, 0.7, 1.3}, {0.15, 0.35, 0.6, 1.3}, {0.66, 0.655, 0.65, 1.3}}, {{0.65, 0.65, 0.65, 1.2}, {0.2, 0.3, 0.6, 1.2}, {0.68, 0.1, 0.11, 1.3}, {0.24, 0.48, 0.48, 1.2}, {0, 0.04, 0.13, 1.2}, {0.3, 0, 0, 1.2}}} --Traffic colors; follow the curly brackets
local zoneData = {{0, 3, 0}, {1, 3, 0}, {2, 3, 5}, {3, 3, 10}, {4, 3, 15}, {5, 3, 20}, {6, 3, 25}, {7, 3, 25}, {8, 3, 25}, {9, 3, 10}, {10, 4, 40}, {11, 4, 40}, {12, 4, 40}, {13, 5, 40}, {14, 4, 40}, {15, 6, 40}, {16, 6, 40}, {17, 6, 40}, {18, 3, 40}, {19, 3, 40}, {20, 3, 40}, {21, 3, 40}, {22, 3, 40}, {23, 3, 40}, {24, 3, 40}, {25, 3, 40}, {26, 3, 35}, {27, 5, 35}, {28, 3, 30}, {29, 3, 30}, {30, 3, 30}, {31, 3, 30}, {32, 3, 30}, {33, 3, 30}, {34, 3, 30}, {35, 3, 25}, {36, 3, 25}, {37, 3, 25}, {38, 3, 25}, {39, 4, 20}, {40, 5, 15}, {41, 5, 10}, {42, 6, 30}, {43, 6, 30}, {44, 6, 30}, {45, 6, 30}, {46, 6, 30}, {47, 6, 30}, {48, 6, 30}, {49, 6, 30}} --Zone data; ID, side offset, forwards probability
local quatAngle = {quatFromDir(vec3(0, 1, 0), vec3(0, 0, 1)), quatFromDir(vec3(1, 0, 0), vec3(0, 0, 1)), quatFromDir(vec3(0, -1, 0), vec3(0, 0, 1)), quatFromDir(vec3(-1, 0, 0), vec3(0, 0, 1))} --Angles; 0, 90, 180, 270

local zoneNames = {}
local player = {}
local traffic = {}
local barriers = {}
local Traffic = {}
local Barrier = {}

local queue = 0
local roadblockZone = 1
local speedLimit = 0
local mainTimer = 0
local hornHold = 0
local funMode = false
local roadblockMode = false

local timeLimit = 200 --Time limit, in seconds
local playerSmash = 48000 --Amount of damage on the player car that will fail the scenario
local policeSmash = 5000 --Amount of damage on a police car that will trigger roadblocks
local trafficSmash = 2000 --Amount of damage on a traffic car that will enable it to reinitialize when spawning
local minSpeed = 15 --Minimum speed limit for randomizer, in metres per second
local maxSpeed = 20 --Maximum speed limit for randomizer, in metres per second
local speedingSpeed = 28 --Excessive speed; can trigger police mode
local crawlSpeed = 4 --Sloth speed; can trigger police arrest
local zoneGap = 3 --Gap between current zone and traffic spawn zone; 3 is optimal
local trafficAggro = 0.9 --Aggression level of the traffic cars
local nearDist = 20 --Near distance, used for traffic passing and interactions
local farDist = 320 --Far distance, used for traffic despawning
local readyDist = 100 --Normal distance, used for a variety of traffic and police logic
local farMultiplier = 3 --Spawn logic multipler for distant traffic cars; used with readyDist
local policeCar = 1 --Which AI vehicle is the police car?
local policeTime = 40 --Police car will be ready after this many seconds
local randomColors = false --If true, traffic car colors can get changed to anything!

function Traffic:new(obj, pos, oldDamage, damage, state, role, travel, zone, far, timer)
	local self = setmetatable({}, Traffic)
	self.__index = self
	self.obj = obj or nil --The vehicle itself
	self.pos = pos or nil --Vehicle position
	self.oldDamage = oldDamage or 0 --Vehicle damage points without player hit
	self.damage = damage or 0 --Vehicle damage points
	self.state = state or nil --Teleport states; none, normal, ready, set
	self.role = role or nil --Purpose of the traffic car; normal, policeNormal, policeChase, roadblock
	self.travel = travel or nil --Direction of travel; none, backward, forward
	self.zone = zone or nil --Zone that the vehicle is currently in
	self.far = far or 1 --Multiplier for spawn mode; see below
	self.timer = timer or 0 --Internal logic timer
	return self
end

function Barrier:new(obj, pos, damage, state, role, zone, timer)
	local self = setmetatable({}, Traffic)
	self.__index = self
	self.obj = obj or nil --The vehicle itself
	self.pos = pos or nil --Vehicle position
	self.damage = damage or 0 --Vehicle damage points
	self.state = state or nil --Teleport states; none, normal, ready, set
	self.role = role or nil --Purpose of the vehicle
	self.zone = zone or nil --Zone that the vehicle is currently in
	self.timer = timer or 0 --Internal logic timer
	return self
end

local function setSpawn(aVehicle) --Sets the new spawn point of a traffic car and rotates it accordingly
	local allow = true
	local randomWay = math.random(1, 100)
	local zoneNumber = player.zone + zoneGap
	local zoneIn = scenetree.findObject(zoneNames[zoneNumber])
	local zonePos = zoneIn:getPosition()
	local zoneRot = zoneIn:getRotation()
	local objAngle = 0
	local posAngle = 0
	local shiftUp = 0
	local shiftOut = 0
	local posOffset = nil
	local rotOffset = nil
	local colorIn = nil
	local colorValues = nil
	if aVehicle.role == "policeNormal" or (aVehicle.role == "policeChase" and roadblockMode == true) then --Normal police always spawn going forwards
		randomWay = 0
	end
	for i, v in ipairs(traffic) do
		if traffic[i].obj.name == aVehicle.obj.name then
			colorIn = paintColors[i][math.random(1, #paintColors[i])]
			shiftUp = 0.5
		end
		if traffic[i].zone == zoneNumber then
			allow = false	
		end
	end
	shiftOut = zoneData[zoneNumber][2]
	if randomWay <= zoneData[zoneNumber][3] then
		objAngle = 3
		posAngle = 4
		if aVehicle.role == "policeChase" and roadblockMode == true then --Police roadblock position
			objAngle = math.random(1, 2) * 2
			shiftOut = 2.8
		end
	else
		objAngle = 1
		posAngle = 2
	end
	if allow == true then
		if randomWay <= zoneData[zoneNumber][3] then
			aVehicle.travel = "forward"
		else
			aVehicle.travel = "backward"
		end
		posOffset = vec3(zonePos) - vec3(0, 1, 0):rotated(zoneRot):rotated(quatAngle[posAngle]) * shiftOut --New position derived from spawn zone position
		rotOffset = quatFromDir(vec3(0, 1, 0):rotated(zoneRot):rotated(quatAngle[objAngle]), vec3(0, 0, 1)) --New rotation derived from spawn zone rotation
		aVehicle.obj:setPositionRotation(posOffset.x, posOffset.y, posOffset.z + shiftUp, 0, 0, rotOffset.z, rotOffset.w + (randomWay / 1000)) --Spawn action; the random addition fixes a spawn rotation init bug
		log("Z", "gull", aVehicle.obj.name.." teleported to "..zoneIn.name)
		if aVehicle.damage >= trafficSmash then
			aVehicle.obj:queueLuaCommand("vehicleResetted(false)")
		end
		if aVehicle.role == "policeChase" and roadblockMode == true then
			aVehicle.role = "roadblock"
			roadblockZone = zoneNumber
		end
		if aVehicle.role == "policeChase" or aVehicle.role == "roadblock" then
			aVehicle.obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
			player.policeWatch = "chase"
		end
		if randomColors == true then
			colorValues = {math.random(), math.random(), math.random(), math.random() + 1}
		else
			colorValues = {colorIn[1], colorIn[2], colorIn[3], colorIn[4]}
		end
		aVehicle.obj.color = ColorF(colorValues[1], colorValues[2], colorValues[3], colorValues[4]):asLinear4F()
		aVehicle.obj.colorPalette0 = aVehicle.obj.color
		aVehicle.state = "set"
		aVehicle.timer = 0
	end
end

local function setPropSpawn(aProp, shiftOut) --Sets the new spawn point of a prop and rotates it accordingly
	if not shiftOut then
		shiftOut = 0
	end
	local zoneNumber = roadblockZone
	local zoneIn = scenetree.findObject(zoneNames[zoneNumber])
	local zonePos = zoneIn:getPosition()
	local zoneRot = zoneIn:getRotation()
	local objAngle = 0
	local posAngle = 0
	local shiftUp = 0
	local posOffset = nil
	local rotOffset = nil
	if aProp.role == "barrier" then
		objAngle = 1
		posAngle = 2
		shiftUp = 0.3
	end
	posOffset = vec3(zonePos) - vec3(0, 1, 0):rotated(zoneRot):rotated(quatAngle[posAngle]) * shiftOut
	rotOffset = quatFromDir(vec3(0, 1, 0):rotated(zoneRot):rotated(quatAngle[objAngle]), vec3(0, 0, 1))
	aProp.obj:setPositionRotation(posOffset.x, posOffset.y, posOffset.z + shiftUp, 0, 0, rotOffset.z, rotOffset.w)
	log("Z", "gull", aProp.obj.name.." teleported to "..zoneIn.name)
	if aProp.damage >= trafficSmash then
		aProp.obj:queueLuaCommand("vehicleResetted(false)")
	end
	aProp.state = "set"
end

local function rushPlayer() --Gives the player a countdown to the time limit
	if player.timer % 1 == 0 then
		local countdown = 10 - math.modf(player.timer)
		if countdown > 0 then
			helper.flashUiMessage((tostring(countdown).."..."), 0.5, true)
		end
	end
	player.timer = player.timer + 0.25
	if player.timer > 10 then
		player.success = false
		scenario.finish({failed = failMessages.overTime})
	end
end

local function arrestPlayer() --Arrests the player with a countdown
	if player.arrestTimer % 1 == 0 then
		local countdown = 3 - math.modf(player.arrestTimer)
		if countdown > 0 then
			helper.flashUiMessage((tostring(countdown).."..."), 0.5, true)
		end
	end
	player.arrestTimer = player.arrestTimer + 0.25
	if player.arrestTimer > 3 then
		player.success = false
		scenario.finish({failed = failMessages.busted})
	end
end

local function chasePlayer(aPolice) --Various logic for while the police car is chasing the player
	local avoid = false
	local chaseDist = readyDist / 2
	local playerDist = (player.pos - aPolice.pos):len()
	if roadblockMode == true then --Police will spawn less or more often depending on these conditions
		aPolice.far = farMultiplier / 1.5
	else
		if player.damage < trafficSmash then --Clean racers get less cops
			aPolice.far = farMultiplier * 1.5
		else
			aPolice.far = farMultiplier
		end
	end
	for i, v in ipairs(traffic) do
		if traffic[i].role == "normal" then
			traffic[i].pos = traffic[i].obj:getPosition()
			local policeDist = (aPolice.pos - traffic[i].pos):len()
			if playerDist <= readyDist and policeDist <= chaseDist then --Traffic reacts to active AI police
				traffic[i].obj:queueLuaCommand("ai.setSpeed(5)")
				if aPolice.travel == "forward" and traffic[i].travel == "forward" then --AI police will try to dodge traffic
					avoid = true
				end
			end
		end
	end
	if playerDist <= chaseDist then
		player.arrestWatch = true
		if (aPolice.travel == "backward" or playerDist <= nearDist) and player.policeWatch == "chase" then
			aPolice.obj:queueLuaCommand("ai.setMode('chase')")
			aPolice.obj:queueLuaCommand("ai.driveInLane('off')")
			player.policeWatch = "force"
		end
		if player.vel <= crawlSpeed then
			arrestPlayer()
		else
			player.arrestTimer = 0
		end
	else
		player.arrestWatch = false
		if aPolice.travel == "backward" and player.policeWatch == "force" then
			aPolice.obj:queueLuaCommand("ai.setMode('random')")
			aPolice.obj:queueLuaCommand("ai.driveInLane('on')")
			player.policeWatch = "chase"
		end
	end
	if avoid == false then
		if player.policeWatch == "chase" then
			aPolice.obj:queueLuaCommand("ai.setSpeed("..tostring(math.max(player.vel * ((readyDist / math.min(readyDist, playerDist) / 10) + 0.5), 5))..")") --Variable police speed based on distance
		end
		if player.policeWatch == "force" then
			aPolice.obj:queueLuaCommand("ai.setSpeed("..tostring(math.max(player.vel * 1.5, 5))..")")
		end
	else
		aPolice.obj:queueLuaCommand("ai.setSpeed("..tostring(speedingSpeed / 1.5)..")")
		aPolice.obj:queueLuaCommand("ai.driveInLane('off')")
	end
end

local function onScenarioUIReady()
	player = {obj = nil, pos = nil, vel = nil, dist = nil, damage = nil, damageTotal = nil, zone = nil, timer = nil, failTimer = nil, policeWatch = nil, arrestWatch = nil, arrestTimer = nil, number = nil, crazy = nil, success = nil}
	player.obj = scenetree.findObject(playerName)
	player.damageTotal = playerSmash
	player.zone = 1
	player.timer = 0
	player.failTimer = 0
	player.policeWatch = "none"
	player.arrestWatch = false
	player.arrestTimer = 0
	player.number = 0
	player.crazy = 0
	player.success = true
	queue = 0
	roadblockZone = 1
	mainTimer = 0
	hornHold = 0
	zoneGap = 3
	roadblockMode = false
	for  i = 1, #zoneData, 1 do
		zoneNames[i] = "spawnzone"..(i - 1)
	end
	for i, v in ipairs(trafficNames) do
		traffic[i] = Traffic:new(scenetree.findObject(trafficNames[i]), nil, 0, 0, "normal", "normal", "none", nil, farMultiplier, 0) --obj, pos, oldDamage, damage, state, role, travel, zone, far, timer
		local zonePos = scenetree.findObject(zoneNames[i]):getPosition()
		local zoneRot = scenetree.findObject(zoneNames[i]):getRotation()
		traffic[i].obj:setPositionRotation(zonePos.x, zonePos.y, zonePos.z + 0.5, 0, 0, zoneRot.z, zoneRot.w)
		traffic[i].travel = "backward"
		if i == policeCar then --AI police car
			traffic[i].state = "none"
			traffic[i].role = "policeNormal"
		end
	end
	for i, v in ipairs(barrierNames) do
		barriers[i] = Barrier:new(scenetree.findObject(barrierNames[i]), nil, 0, "normal", "barrier", nil, 0) --obj, pos, damage, state, role, zone, timer
	end
end

local function onRaceStart()
	player.obj = scenetree.findObject(playerName) --?
	local policeVehicles = {"vehicles/coupe/police.pc", "vehicles/super/polizia.pc", "vehicles/etk800/police.pc", "vehicles/fullsize/police.pc", "vehicles/fullsize/bcpd.pc", "vehicles/etkc/police_highway_A.pc", "vehicles/autobello/carabinieri.pc", "vehicles/burnside/police.pc", "vehicles/sunburst/police.pc", "vehicles/sunburst/police_RS.pc", "vehicles/van/police.pc", "vehicles/roamer/sheriff.pc", "vehicles/roamer/sheriff_offroad.pc"}
	local crazyVehicles = {"ball", "barrels", "barrier", "blockwall", "bollard", "boxutility", "boxutility_large", "cannon", "caravan", "christmas_tree", "citybus", "cones", "dryvan", "flail", "flatbed", "flipramp", "haybale", "inflated_mat", "kickplate", "large_angletester", "large_bridge", "large_cannon", "large_crusher", "large_hamster_wheel", "large_metal_ramp", "large_roller", "large_spinner", "large_tilt", "metal_box", "metal_ramp", "piano", "pigeon", "roadsigns", "rocks", "rollover", "sawhorse", "semi", "streetlight", "suspensionbridge", "tanker", "tirestacks", "tirewall", "trafficbarrel", "tsfb", "tube", "van", "wall", "weightpad", "woodcrate", "woodplanks"}
	local ramplow = "vehicles/semi/t65_ramplow.pc"
	player.number = 1
	for i, v in ipairs(policeVehicles) do
		if player.obj.partConfig == policeVehicles[i] then
			player.crazy = 1
			player.number = 1.5 --Modifies speed limit, police start spawn time, police speed watch
		end
	end
	for i, v in ipairs(crazyVehicles) do
		if player.obj.jbeam == crazyVehicles[i] and player.crazy ~= 1 then
			player.crazy = 2
			player.number = 1.1
		end
	end
	if player.obj.partConfig == ramplow then
		player.crazy = 3
		player.number = 0.6
	end
	speedLimit = math.ceil(math.random(minSpeed, maxSpeed) / player.number) --Creates a new speed limit for all traffic cars; makes restarts more interesting but prevents traffic cars from bumping into each other
	log("Z", "gull", "Speed limit: "..speedLimit)
	for i, v in ipairs(traffic) do
		traffic[i].obj:queueLuaCommand("ai.setAggression("..tostring(trafficAggro)..")")
		traffic[i].obj:queueLuaCommand("ai.driveInLane('on')")
		traffic[i].obj:queueLuaCommand("ai.setSpeed("..tostring(speedLimit)..")")
		traffic[i].obj:queueLuaCommand("ai.setSpeedMode('limit')")
		traffic[i].obj:queueLuaCommand("ai.resetLearning(true)")
		traffic[i].obj:queueLuaCommand("ai.setMode('random')")
	end
end

local function onRaceTick(raceTickTime)
	mainTimer = mainTimer + raceTickTime
	player.pos = player.obj:getPosition()
	player.vel = map.objects[map.objectNames[player.obj.name]].vel:length()
	player.damage = map.objects[map.objectNames[player.obj.name]].damage
	if mainTimer < 1 then --Workaround for an odd bug with zWeight where it won't get the proper value the first time
		player.obj:queueLuaCommand("obj:queueGameEngineLua('zWeight = ('..obj:calcBeamStats().total_weight..')')") --Creates a global variable with the player vehicle weight
	elseif mainTimer == 1 then
		if zWeight >= 2000 and funMode == false then
			player.damageTotal = math.ceil(player.damageTotal * (1 + (zWeight / 5000))) --Extra damage tolerance formula
		end
		log("Z", "gull", "Damage limit: "..player.damageTotal)
	end
	player.obj:queueLuaCommand("obj:queueGameEngineLua('zHorn = ('..electrics.values.horn..')')") --Creates a global variable for the horn state; 0 = off and 1 = on
	if zHorn == 1 then
		hornHold = hornHold + raceTickTime
		if hornHold == 1 and mainTimer <= 10 then
			timeLimit = 10000
			playerSmash = 1000000
			player.damageTotal = playerSmash
			funMode = true
			helper.flashUiMessage("Fun mode enabled!", 2, false)
		end
	end
	for i, v in ipairs(traffic) do
		traffic[i].pos = traffic[i].obj:getPosition()
		player.dist = (player.pos - traffic[i].pos):len()
		traffic[i].damage = map.objects[map.objectNames[traffic[i].obj.name]].damage
		if traffic[i].state == "set" then
			if traffic[i].timer >= 0.5 then --Any traffic timers, such as here, intentionally delay AI mode changes; experimental!
				if traffic[i].role == "roadblock" then
					traffic[i].obj:queueLuaCommand("ai.setMode('stopping')")
					barriers[1].timer = 0
				else
					traffic[i].obj:queueLuaCommand("ai.setMode('random')")
				end
				traffic[i].obj:queueLuaCommand("ai.driveInLane('on')")
				traffic[i].state = "normal"
				traffic[i].far = 1
			end
			traffic[i].timer = traffic[i].timer + raceTickTime
		end
		if traffic[i].role == "roadblock" and (traffic[i].state == "normal" or traffic[i].state == "ready") then
			if barriers[1].timer < #barriers then --Uses barrier timer to delay barrier spawns for slight lag reduction
				for o, v in ipairs(barriers) do
					if (barriers[1].timer + 1) % 1 == 0 then
						if barriers[o].obj.name == barriers[barriers[1].timer + 1].obj.name then
							setPropSpawn(barriers[o], (o * 3.5) - 1.5)
						end
					end
				end
				barriers[1].timer = barriers[1].timer + raceTickTime
			end
		end
		if mainTimer >= policeTime * player.number then
			if traffic[i].role == "policeNormal" and player.policeWatch == "none" then
				traffic[i].state = "normal"
				player.policeWatch = "patrol"
			end
		end
		if player.dist <= nearDist then --If you pass a traffic car, prepare it for teleportation
			traffic[i].state = "ready"
			traffic[i].far = 1
			if traffic[i].role == "policeNormal" then
				traffic[i].far = 2
			end
			traffic[i].timer = 0
		end
		if player.dist <= readyDist / 2 then --A bunch of near area logic
			if traffic[i].role == "policeNormal" and player.policeWatch == "patrol" and (player.vel >= speedingSpeed * player.number or (player.dist <= 10 and traffic[i].damage >= traffic[i].oldDamage + (trafficSmash / 2))) then
				traffic[i].obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
				traffic[i].obj:queueLuaCommand("ai.setAggression(1.2)")
				traffic[i].role = "policeChase"
				player.policeWatch = "chase"
			end
			if traffic[i].role == "policeChase" and player.dist <= 10 and traffic[i].damage >= policeSmash and roadblockMode == false then
				helper.flashUiMessage("The cops are really mad now!", 3, false)
				player.damageTotal = player.damageTotal * 1.5
				roadblockMode = true
			end
			if traffic[i].role == "roadblock" and player.dist <= nearDist then
				traffic[i].role = "policeChase"
			end
			traffic[i].oldDamage = traffic[i].damage
		end
		if player.zone > 1 then --Enables traffic spawning and other behaviors after you pass enough zones
			if player.dist >= farDist and traffic[i].state == "normal" then --If you get really far away from a traffic car, prepare it for teleportation
				if traffic[i].role ~= "policeChase" then
					traffic[i].far = farMultiplier --Multiplier for ready distance; can prevent sudden vanishing
				end
				traffic[i].state = "ready"
				traffic[i].timer = 0
			end
			if queue > 0 then
				if player.dist >= (readyDist * traffic[i].far) and traffic[i].state == "ready" then
					if traffic[i].timer == 0 and traffic[i].zone == "none" then
						traffic[i].obj:queueLuaCommand("ai.setMode('disabled')")
						traffic[i].far = 0
					end
					if player.dist >= readyDist and traffic[i].timer >= 0.5 and traffic[i].obj.name == traffic[queue].obj.name then
						setSpawn(traffic[i])
						queue = -4 --Artificial delay for next spawn test; prevents lag and stuff
					end
					traffic[i].timer = traffic[i].timer + raceTickTime
				end
			end
		end
		if traffic[i].role == "policeChase" then --Function always runs if the police is chasing you
			chasePlayer(traffic[i])
		end
		if traffic[i].role ~= "policeChase" and traffic[i].damage >= trafficSmash * 2 then --Traffic cars try to stop driving if heavily damaged
			traffic[i].obj:queueLuaCommand("ai.setSpeed("..tostring(trafficSmash * 8 / traffic[i].damage)..")")
		end
	end
	for i, v in ipairs(barriers) do
		barriers[i].damage = map.objects[map.objectNames[barriers[i].obj.name]].damage
	end
	queue = queue + 1 --This gets updated every race tick to prevent two cars from being teleported at the same time to the same spawn zone
	if queue > #trafficNames then
		queue = 1
	end
	if player.damage >= player.damageTotal then
		player.failTimer = player.failTimer + raceTickTime
		if player.success == true then
			bullettime.set(0.5)
			player.success = false
		end
		if player.failTimer >= 2 then
			if player.arrestWatch == true then
				scenario.finish({failed = failMessages.crashBusted})
			else
				scenario.finish({failed = failMessages.crash})
			end
		end
	end
	if timeLimit - 10 <= mainTimer then
		rushPlayer()
	end
end

local function onRaceResult()
	if player.success == true then
		if player.crazy == 1 then --Chaotic good
			scenario.finish({msg = winMessages.policeCar})
		elseif player.crazy == 2 then --Chaotic neutral
			scenario.finish({msg = winMessages.funnyCar})
		elseif player.crazy == 3 then --Chaotic evil
			scenario.finish({msg = winMessages.evilCar})
		elseif funMode == true then --Cheaty
			scenario.finish({msg = winMessages.normal})
		elseif mainTimer <= timeLimit - 60 then --Lawful evil
			scenario.finish({msg = winMessages.ultraFast})
		elseif mainTimer <= timeLimit - 40 then --Lawful neutral
			scenario.finish({msg = winMessages.veryFast})
		elseif mainTimer >= timeLimit - 5 then --Neutral good
			scenario.finish({msg = winMessages.almost})
		elseif roadblockZone > 1 then --Neutral evil
			scenario.finish({msg = winMessages.lawless})
		elseif player.policeWatch == "patrol" then --Lawful good
			scenario.finish({msg = winMessages.lawful})
		else --True neutral
			scenario.finish({msg = winMessages.normal})
		end
	end
end

local function onBeamNGTrigger(data)
	if data.event == "enter" then
		if data.subjectName == player.obj.name then 
			if string.byte(tostring(data.triggerName)) == 115 then --Prevents stupid trigger confusion with bus stops
				for i, v in ipairs(zoneNames) do
					if data.triggerName == zoneNames[i] then
						if player.zone > i and i >= math.abs(zoneGap) + 1 then
							zoneGap = -3
						else
							zoneGap = 3
						end
						player.zone = i
					end
				end
				if player.zone + zoneGap > #zoneData then --Prevents spawn zone errors
					player.zone = 0
				end
				log("Z", "gull", "Zone check: "..player.zone)
			end
			if string.byte(tostring(data.triggerName)) == 102 then --Fail zones
				player.success = false
				if data.triggerName == "failzone0" then
					scenario.finish({failed = outMessages.roadClosed})
				end
				if data.triggerName == "failzone1" then
					scenario.finish({failed = outMessages.wrongWay})
				end
				if data.triggerName == "failzone2" then
					scenario.finish({failed = outMessages.hospital})
				end
			end
		end
		for i, v in ipairs(traffic) do
			if data.subjectName == traffic[i].obj.name then
				if string.byte(tostring(data.triggerName)) == 105 then --Intersections
					if traffic[i].role ~= "policeChase" then
						traffic[i].obj:queueLuaCommand("ai.setSpeed(8)")
					end
				end
				if string.byte(tostring(data.triggerName)) == 97 then --Action zones
					traffic[i].obj:queueLuaCommand("ai.setSpeed(0.5)")
				end
				for o, v in ipairs(zoneNames) do
					if data.triggerName == zoneNames[o] then
						traffic[i].zone = o
					end
				end	
				if data.triggerName == zoneNames[player.zone + zoneGap] then
					if traffic[i].role ~= "policeChase" then
						traffic[i].obj:queueLuaCommand("ai.setSpeed(10)") --Traffic cars slow down in zones where teleports are likely
					end
				end
				if data.triggerName == zoneNames[roadblockZone] then
					if traffic[i].role ~= "policeChase" then
						traffic[i].obj:queueLuaCommand("ai.setSpeed(0.5)") --Traffic cars slow down by police roadblocks
					end
				end
				log("Z", "gull", traffic[i].obj.name.." entered zone "..data.triggerName)
			end
		end
	end
	if data.event == "exit" then
		for i, v in ipairs(traffic) do
			if data.subjectName == traffic[i].obj.name then
				if string.byte(tostring(data.triggerName)) == 115 then
					if traffic[i].state ~= "set" then
						traffic[i].zone = "none"
					end
					if traffic[i].role ~= "policeChase" then
						traffic[i].obj:queueLuaCommand("ai.setSpeed("..tostring(speedLimit)..")")
					end
				end
				log("Z", "gull", traffic[i].obj.name.." exited zone "..data.triggerName)
			end
		end
	end
end

M.onScenarioUIReady = onScenarioUIReady
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onRaceResult = onRaceResult
M.onBeamNGTrigger = onBeamNGTrigger

return M